/*
 * Direct3D GLX interface
 *
 * Copyright 2001 TransGaming Technologies, Inc.
 */

#include "config.h"

#include "ts_xlib.h"
#include "x11drv.h"
#include "x11font.h"
#include "x11ddraw.h"
#include "glx.h"
#include "executebuffer.h"
#include "d3d.h"
#include "d3dhal.h"
#include "wine_gl.h"
#include "wine/d3dhalgl.h"

#include "windef.h"
#include "wingdi.h"
#include "ddrawi.h"
#include "bitmap.h"
#include "win.h"
#include "debugtools.h"
#include "options.h"
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

DEFAULT_DEBUG_CHANNEL(x11drv);

#define RENDER_TO_WINDOW
/* #define TRANSGAMING_LOGO */

DWORD X11DRV_GLX_TLS;

static BOOL glx_available = FALSE;

void X11DRV_GLX_Init(void)
{
  int glx_event, glx_error, glx_major, glx_minor;

  if (glx_available) return; /* already initialized? */

  ENTER_GL();
  if (glXQueryExtension(gdi_display, &glx_event, &glx_error) &&
      glXQueryVersion(gdi_display, &glx_major, &glx_minor)) {

    TRACE("GLX extension detected and enabled\n");
    glx_available = TRUE;

    X11DRV_GLX_TLS = TlsAlloc();

    X11DRV_GLX_EB_Init();
  }
  LEAVE_GL();
}

void X11DRV_GLX_Cleanup(void)
{
  X11DRV_GLX_EB_Cleanup();
  if (TlsFree(X11DRV_GLX_TLS))
    X11DRV_GLX_TLS = 0;
}

/* actual Direct3D HAL stuff */

typedef struct _D3DCONTEXT {
  GLXContext	ctx;
  XVisualInfo*	vis;
  GLXDrawable	draw, ldraw;
  DWORD		flags;
  pid_t		owner_pid;
  HBITMAP	bmp, lbmp, zbmp;
  BITMAPOBJ*	lockedBmp;
  BITMAPOBJ*	lockedZBmp;
  X11DRV_DIBSECTION *dib, *ldib, *zdib;
  LPDDRAWI_DDRAWSURFACE_LCL lsurf, zsurf, nsurf;
  LPDDRAWI_DDRAWSURFACE_GBL surf;
  Font		lfont;

  GLint		texunits;
  void (*glActiveTextureARB)(GLenum texture);

#ifdef TRANSGAMING_LOGO
  GLuint logo;
  GLuint logo_list;
  DWORD logo_size;
#endif

  GLXContext	save_ctx;
  GLXDrawable	save_draw;

  struct _D3DCONTEXT *next;
  struct _D3DCONTEXT *prev;
} D3DCONTEXT,*LPD3DCONTEXT;

#define FLAG_PIXMAP	0x01
#define FLAG_CINDIRECT	0x02
#define FLAG_PINDIRECT	0x04
#define FLAG_WINDIRECT  0x08
#define FLAG_PDIRECT	0x10
#define FLAG_WDIRECT	0x20
#define FLAG_INDIRECT	(FLAG_PINDIRECT|FLAG_WINDIRECT)
#define FLAG_DIRECT	(FLAG_PDIRECT|FLAG_WDIRECT)
#define FLAG_BACKBUF	0x40
#define FLAG_FLIPPED	0x80

static DWORD X11DRV_GLX_MakeCurrent(LPD3DCONTEXT ctx);
static void X11DRV_GLX_KillCurrent(LPD3DCONTEXT ctx);

static void X11DRV_GLX_SetupLogo(LPD3DCONTEXT ctx);
static void X11DRV_GLX_RenderLogo(LPD3DCONTEXT ctx);

static inline LPD3DCONTEXT X11DRV_GLX_GetCurrent(void)
{
  return TlsGetValue(X11DRV_GLX_TLS);
}

static inline void X11DRV_GLX_SetCurrent(LPD3DCONTEXT ctx)
{
  TlsSetValue(X11DRV_GLX_TLS, ctx);
}

static int GLXError;
static int GLXErrorHandler(Display *dpy, XErrorEvent *event)
{
  GLXError = 1;
  return 0;
}

static void X11DRV_GLX_Format(LPDIBSECTION dib, GLenum *fmt, GLenum *type)
{
  /* FIXME: just guessing */
  switch (dib->dsBmih.biBitCount) {
  case 15:
    *fmt=GL_RGBA; *type=GL_UNSIGNED_SHORT_5_5_5_1; /* or something */
    break;
  case 16:
    *fmt=GL_RGB; *type=GL_UNSIGNED_SHORT_5_6_5;
    break;
  case 24:
    *fmt=GL_RGB; *type=GL_UNSIGNED_BYTE;
    break;
  case 32:
    *fmt=GL_RGBA; *type=GL_UNSIGNED_INT_8_8_8_8;
    break;
  default:
    ERR("unhandled DIB depth %d\n", dib->dsBmih.biBitCount);
    *fmt=0; *type=0;
  }
}

static void X11DRV_GLX_ZFormat(LPDIBSECTION dib, GLenum *fmt, GLenum *type)
{
  *fmt = GL_DEPTH_COMPONENT;
  switch (dib->dsBmih.biBitCount) {
  case 8:
    *type = GL_UNSIGNED_BYTE;
    break;
  case 15:
  case 16:
    *type = GL_UNSIGNED_SHORT;
    break;
  case 24:
  case 32:
    *type = GL_UNSIGNED_INT;
    break;
  default:
    ERR("unhandled DIB depth %d\n", dib->dsBmih.biBitCount);
    *fmt=0; *type=0;
  }
}

static void X11DRV_GLX_Destroy(LPD3DCONTEXT ctx)
{
  if (!(ctx->flags & FLAG_PIXMAP)) {
    /* get rid of any AuxMod */
    if (ctx->bmp) X11DRV_CoerceDIBSection2(ctx->bmp, DIB_Status_InSync, FALSE);
    if (ctx->lbmp) X11DRV_CoerceDIBSection2(ctx->lbmp, DIB_Status_InSync, FALSE);
  }
  if (ctx->zbmp) X11DRV_CoerceDIBSection2(ctx->bmp, DIB_Status_InSync, FALSE);
  if (ctx->ctx) {
    X11DRV_GLX_KillCurrent(ctx);
    glXDestroyContext(gdi_display, ctx->ctx);
  }
  if (ctx->flags & FLAG_PIXMAP) {
    glXDestroyGLXPixmap(gdi_display, ctx->draw);
    if (ctx->lbmp) glXDestroyGLXPixmap(gdi_display, ctx->ldraw);
  }
  XFree(ctx->vis);
}

/* This function is called with GLX resources locked */
static void X11DRV_GLX_CopyDIB(LPD3DCONTEXT ctx, BOOL copyZbmp)
{
  BITMAPOBJ *bmp;
  X11DRV_DIBSECTION *dib;
  GLenum fmt, type;
  unsigned w, h, p;
  LPBYTE bits;
  int i;

  TRACE("copying from %s DIBsection to GL buffer\n", copyZbmp?"Zbuffer":"RGB");

  /* First check if we're copying from the zbmp or the pixel bmp */
  if (copyZbmp)
    bmp = ctx->lockedZBmp;
  else
    bmp = ctx->lockedBmp;
  if (!bmp)
    return;

  dib = (X11DRV_DIBSECTION *)bmp->dib;

  /* make the DIB readable for us */
  if (dib->status == DIB_Status_GdiMod) {
    TRACE("state is GdiMod, converting to InSync\n");
    glXWaitGL();
    X11DRV_DIB_Coerce(bmp, DIB_Status_InSync, FALSE);
    dib->p_status = DIB_Status_InSync;
    glXWaitX();
  }
  w = dib->dibSection.dsBm.bmWidth;
  h = dib->dibSection.dsBm.bmHeight;
  p = dib->dibSection.dsBm.bmWidthBytes;
  bits = dib->dibSection.dsBm.bmBits;
  if (dib == ctx->zdib) {
    X11DRV_GLX_ZFormat(&dib->dibSection, &fmt, &type);
  } else {
    X11DRV_GLX_Format(&dib->dibSection, &fmt, &type);
  }
  TRACE("fmt=%x, type=%x, w=%d, h=%d, bits=%p, state=%d\n", fmt, type, w, h, bits, dib->status);

  /* Save GL state info.
   * glPixelZoom changes stuff in GL_PIXEL_MODE_BIT, but nobody else calls
   * glDrawPixels or glCopyPixels yet. */
  glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT |
               GL_DEPTH_BUFFER_BIT | GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);
 
  /* Set up the transformation so that we can set the rasterpos appropriately */
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0.0, w, h, 0.0, -1, 1);

  /* Disable OpenGL features that might slow down glDrawPixels. */
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_DITHER);
  glDisable(GL_FOG);
  glDisable(GL_LIGHTING);
  glDisable(GL_SCISSOR_TEST);
  glDisable(GL_STENCIL_TEST);
  if (ctx->texunits > 1) {
    for (i=0; i<ctx->texunits; i++) {
      ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
      glDisable(GL_TEXTURE_2D);
    }
  }
  else glDisable(GL_TEXTURE_2D);

  if (fmt == GL_DEPTH_COMPONENT) {
    glDepthFunc(GL_ALWAYS);
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
  } else {
    glDisable(GL_DEPTH_TEST);
  }

  /* Transfer pixels. */
  if (fmt == GL_DEPTH_COMPONENT)
  {
      /* This is much faster on nVidia and Mesa-3.5.
       * Don't know about others. */
      DWORD y;

      glViewport(0, 0, ctx->lsurf->lpGbl->wWidth, ctx->lsurf->lpGbl->wHeight);

      for (y=0; y < h; y++)
      {
	  glPixelZoom(1, 1);
	  glRasterPos2i(0, y);
	  glDrawPixels(w, 1, fmt, type, bits);
	  bits += p;
      }
  }
  else
  {
      /* We need Y=1 as a side-effect of using glPixelZoom. */
      glViewport(0, 1, ctx->lsurf->lpGbl->wWidth, ctx->lsurf->lpGbl->wHeight);

      glPixelZoom(1,-1);
      glRasterPos2i(0,0);
      glDrawPixels(w, h, fmt, type, bits);
  }

  /* Restore GL state info and finish */
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();

  dib->status = DIB_Status_AuxMod;
}

static void X11DRV_GLX_copy_aux(BITMAPOBJ*bmp, void*c, int req)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)c;
  X11DRV_DIBSECTION *dib = (X11DRV_DIBSECTION *)bmp->dib;
  GLenum fmt, type;
  unsigned y, w, h, p;
  LPBYTE bits;
  DWORD back;

  TRACE("copying from GL buffer to DIBsection\n");
  /* because OpenGL's coordinates are upside-down relative to the DIBs DirectDraw uses,
   * we must read out pixel data line by line into app memory */
  X11DRV_DIB_Coerce(bmp, DIB_Status_AppMod, TRUE);
  w = dib->dibSection.dsBm.bmWidth;
  h = dib->dibSection.dsBm.bmHeight;
  p = dib->dibSection.dsBm.bmWidthBytes;
  bits = dib->dibSection.dsBm.bmBits;
  back = ctx->flags & FLAG_BACKBUF;
  if (dib == ctx->zdib) {
    X11DRV_GLX_ZFormat(&dib->dibSection, &fmt, &type);
  } else {
    X11DRV_GLX_Format(&dib->dibSection, &fmt, &type);
    if (dib == ctx->ldib)
      /* hm... this is not the last surface rendered to, but the one before it...
       * let's assume that it's just double-buffering so it's ok to just flip this */
      back ^= FLAG_BACKBUF;
    /* check whether we have done a GLX flip since the last rendering, too */
    if (ctx->flags & FLAG_FLIPPED)
      back ^= FLAG_BACKBUF;
  }
  ENTER_GL();

  /* should we set the context? */
  if (SUCCEEDED(X11DRV_GLX_MakeCurrent(ctx))) {
    glXWaitX();

    glFinish();
    glReadBuffer(back ? GL_BACK : GL_FRONT);
    for (y=0; y<h; y++, bits+=p) {
      glReadPixels(0, h-y, w, 1, fmt, type, bits);
    }
  }
  else {
    ERR("failed to update DIB section from OpenGL frame buffer\n");
  }
  LEAVE_GL();
}

#define FONT_BASE 4200

BOOL X11DRV_GLX_write_aux(BITMAPOBJ *bmp, void*c, DC *dc, INT x, INT y, UINT flags,
                          const RECT *lprect, LPCWSTR wstr, UINT count,
                          const INT *lpDx)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)c;
  X11DRV_PDEVICE *physDev = (X11DRV_PDEVICE *)dc->physDev;
  fontObject *pfo = XFONT_GetFontObject( physDev->font );
  XFontStruct *font = pfo->fs;
  int i;
  GLint rp[2];

  TRACE("hdc=%04x df=%04x %d,%d %s, %d  flags=%d lpDx=%p\n",
        dc->hSelf, (UINT16)(physDev->font), x, y,
        debugstr_wn (wstr, count), count, flags, lpDx);
  /* if an app draws onto a surface used for GLX rendering, it's probably
   * not very advanced text, we'll only handle common cases */
  /* hmm, we ought to be able to handle ETO_OPAQUE and ETO_CLIPPED? */
  if (flags) goto write_fallback;
  if (lpDx) goto write_fallback;
  /* we're not Unicode-aware yet */
  for (i=0; i<count; i++) if (wstr[i] >= 256) goto write_fallback;

  rp[0] = INTERNAL_XWPTODP( dc, x, y );
  rp[1] = INTERNAL_YWPTODP( dc, x, y ) + font->ascent+1;;

  TRACE("pos: %d, %d, color: %08lx\n", rp[0], rp[1], dc->textColor);

  ENTER_GL();
  /* should we set the context? */
  if (SUCCEEDED(X11DRV_GLX_MakeCurrent(ctx))) {
    glXWaitX();

    if (font->fid != ctx->lfont) {
      TRACE("generating gdi_display lists for font %ld\n", font->fid);
      glXUseXFont(font->fid, 0, 256, FONT_BASE);
      ctx->lfont = font->fid;
    }

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT |
                 GL_LIST_BIT | GL_TRANSFORM_BIT);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.0, ctx->lsurf->lpGbl->wWidth, ctx->lsurf->lpGbl->wHeight, 0.0, -1, 1);

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_STENCIL_TEST);
    if (ctx->texunits > 1) {
      for (i=0; i<ctx->texunits; i++) {
        ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
        glDisable(GL_TEXTURE_2D);
      }
    }
    else glDisable(GL_TEXTURE_2D);

    glColor3ub(GetRValue(dc->textColor),
               GetGValue(dc->textColor),
               GetBValue(dc->textColor));
    glRasterPos2iv(rp);

    glListBase(FONT_BASE);
    glCallLists(count, GL_UNSIGNED_SHORT, wstr);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glPopAttrib();

    glXWaitGL();
    LEAVE_GL();
  }
  else {
    LEAVE_GL();
    goto write_fallback;
  }

  return TRUE;

write_fallback:
  return X11DRV_ExtTextOut(dc, x, y, flags, lprect, wstr, count, lpDx);
}

static X11DRV_DIBSECTION *X11DRV_GLX_AttachAux(LPD3DCONTEXT ctx, HBITMAP hBmp)
{
   BITMAPOBJ *bmp;
   X11DRV_DIBSECTION *dib;

   /* attach aux handlers to DIB sections
    * (hoping no more than one context uses it) */
   bmp = (BITMAPOBJ *)GDI_GetObjPtr(hBmp, BITMAP_MAGIC);
   dib = (X11DRV_DIBSECTION *)bmp->dib;
   if (dib) {
     dib->aux_ctx = ctx;
     dib->copy_aux = X11DRV_GLX_copy_aux;
     dib->write_aux = X11DRV_GLX_write_aux;
   }
   GDI_ReleaseObj(hBmp);
   TRACE("aux handler attached to %p (%p)\n", bmp, dib);
   return dib;
}

static void X11DRV_GLX_MakeZbuffer(LPD3DCONTEXT ctx, LPDDRAWI_DDRAWSURFACE_LCL lsurf)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  if (hBmp) {
    X11DRV_DIBSECTION *dib;
    TRACE("setting surf %p to ctx %p\n", lsurf, ctx);
    dib = X11DRV_GLX_AttachAux(ctx, hBmp);
    if (ctx->zbmp) X11DRV_CoerceDIBSection2(ctx->zbmp, DIB_Status_InSync, FALSE);
    ctx->zdib = dib;
    ctx->zbmp = hBmp;
  }
  ctx->zsurf = lsurf;
}

static void X11DRV_GLX_LockTarget(LPDDRAWI_DDRAWSURFACE_LCL lsurf, BOOL lock)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  if (!hBmp) return;
  if (lock) GDI_GetObjPtr( hBmp, BITMAP_MAGIC );
  else GDI_ReleaseObj( hBmp );
}

/* OK, this procedure is getting unmaintainable... */

static void X11DRV_GLX_MakeTarget(LPD3DCONTEXT ctx, LPDDRAWI_DDRAWSURFACE_LCL lsurf)
{
  LPDDRAWI_DDRAWSURFACE_GBL surf = lsurf ? lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
  HBITMAP hBmp = surf ? GET_LPDDRAWSURFACE_GBL_MORE(surf)->hKernelSurface : (HBITMAP)0;
  Drawable draw = 0;

  ENTER_GL();
  if (hBmp) {
    /* save this context in surface (hoping only one context uses it) */
    lsurf->dwReserved1 = (DWORD)ctx;
    TRACE("setting surf %p to ctx %p\n", lsurf, ctx);
#ifdef RENDER_TO_WINDOW
    if (lsurf->ddsCaps.dwCaps & (DDSCAPS_PRIMARYSURFACE | DDSCAPS_BACKBUFFER)) {
      if (ctx->flags & FLAG_PIXMAP) {
        /* destroy context if it's indirect for pixmaps only */
        if (ctx->ctx && ((ctx->flags & FLAG_INDIRECT) == FLAG_PINDIRECT)) {
          X11DRV_GLX_KillCurrent(ctx);
          glXDestroyContext(gdi_display, ctx->ctx);
          ctx->ctx = 0;
          ctx->flags &= ~FLAG_CINDIRECT;
        }
        /* evict pixmaps */
        if (ctx->lbmp) {
          glXDestroyGLXPixmap(gdi_display, ctx->ldraw);
          ctx->ldraw = None;
          ctx->lbmp = (HBITMAP)0;
        }
        glXDestroyGLXPixmap(gdi_display, ctx->draw);
        ctx->draw = None;
        ctx->bmp = (HBITMAP)0;
        ctx->flags &= ~FLAG_PIXMAP;
      }
      if (lsurf != ctx->lsurf) {
        HWND hwnd;

        /* find the drawable for the window that ddraw uses */
        hwnd = (HWND)lsurf->lpSurfMore->lpDDRAWReserved;
        draw = X11DRV_get_client_window(hwnd);
        TRACE("WND: %x Draw: %ld\n", hwnd, draw);
        if (!draw) {
          TRACE("no drawable, host off desktop\n");
          draw = X11DRV_get_client_window(GetDesktopWindow());
          TRACE("Owner WND: %x Drawable: %ld\n", hwnd, draw);
        }
        ctx->draw = draw;
      }
      /* manage the "shadow" DIB section */
      if (hBmp == ctx->lbmp) {
        X11DRV_DIBSECTION *dib = ctx->dib;
        ctx->dib = ctx->ldib;
        ctx->ldib = dib;
        ctx->lbmp = ctx->bmp;
        ctx->bmp = hBmp;
      }
      else if (hBmp != ctx->bmp) {
        X11DRV_DIBSECTION *dib = X11DRV_GLX_AttachAux(ctx, hBmp);
        /* cache previous DIB section */
        if (ctx->lbmp) X11DRV_CoerceDIBSection2(ctx->lbmp, DIB_Status_InSync, FALSE);
        ctx->ldib = ctx->dib;
        ctx->lbmp = ctx->bmp;
        ctx->dib = dib;
        ctx->bmp = hBmp;
      }
      if (lsurf->ddsCaps.dwCaps & DDSCAPS_PRIMARYSURFACE)
        ctx->flags &= ~FLAG_BACKBUF;
      else
        ctx->flags |= FLAG_BACKBUF;
      if (ctx->flags & FLAG_WINDIRECT) ctx->flags |= FLAG_CINDIRECT;
    }
    else
#endif
    if (hBmp == ctx->lbmp) {
      /* flip to cached pixmap */
      GLXDrawable draw = ctx->draw;
      ctx->draw = ctx->ldraw;
      ctx->ldraw = draw;
      ctx->lbmp = ctx->bmp;
      ctx->bmp = hBmp;
    }
    else if (hBmp != ctx->bmp) {
      BITMAPOBJ *bmp;
      bmp = (BITMAPOBJ *)GDI_GetObjPtr( hBmp, BITMAP_MAGIC );
      if (!bmp->physBitmap)
        X11DRV_CreateBitmap(hBmp);
      draw = (Pixmap)bmp->physBitmap;
      GDI_ReleaseObj( hBmp );
      if (ctx->lbmp) {
        /* evict old cached pixmap */
        if (ctx->flags & FLAG_PIXMAP)
          glXDestroyGLXPixmap(gdi_display, ctx->ldraw);
        else
          X11DRV_CoerceDIBSection2(ctx->lbmp, DIB_Status_InSync, FALSE);
        ctx->ldraw = None;
        ctx->lbmp = (HBITMAP)0;
      }
      if (ctx->flags & FLAG_PIXMAP) {
        /* cache previous pixmap, in case the app is double buffering them */
        ctx->ldraw = ctx->draw;
        ctx->lbmp = ctx->bmp;
      } else {
        /* destroy context if it's direct for windows only */
        if (ctx->ctx && ((ctx->flags & FLAG_DIRECT) == FLAG_WDIRECT)) {
          if (ctx->bmp) X11DRV_CoerceDIBSection2(ctx->bmp, DIB_Status_InSync, FALSE);
          X11DRV_GLX_KillCurrent(ctx);
          glXDestroyContext(gdi_display, ctx->ctx);
          ctx->ctx = 0;
        }
        /* need to render to "front" of pixmap */
        ctx->flags &= ~FLAG_BACKBUF;
      }
      /* prepare new pixmap */
      ctx->draw = glXCreateGLXPixmap(gdi_display, ctx->vis, draw);
      ctx->flags |= FLAG_PIXMAP;
      ctx->bmp = hBmp;
      if (ctx->flags & FLAG_PINDIRECT) ctx->flags |= FLAG_CINDIRECT;
    }
  } else {
    /* this is probably GL_DirectDraw_Init trying to determine device caps,
     * it's probably OK to use the root window for this purpose, but the context
     * needs to use the root_windows's visual, not the standard one.  Some GL drivers
     * (ie: NVidia) choke if you have mismatched context/window visuals.  */
    XWindowAttributes root_win_attr;
    Visual *rootVisual;
    XVisualInfo template;
    int num;

    ctx->draw = root_window;
    ctx->flags = 0;
    ctx->bmp = (HBITMAP)0;

    if (XGetWindowAttributes( gdi_display, root_window, &root_win_attr ))
    {
      rootVisual = root_win_attr.visual; 
    }
    else
    {
      /* Get the default visual, since we can't seem to get the attributes from the 
         Root Window.  Let's hope that the Root Window Visual matches the DefaultVisual */
      rootVisual = DefaultVisual( gdi_display, DefaultScreen(gdi_display) );
    }     

    template.visualid = XVisualIDFromVisual(rootVisual);
    XFree(ctx->vis);
    ctx->vis = XGetVisualInfo(gdi_display, VisualIDMask, &template, &num);
    
  }
  ctx->nsurf = lsurf;
  ctx->flags &= ~FLAG_FLIPPED;

  LEAVE_GL();
}

static DWORD X11DRV_GLX_MakeCurrent(LPD3DCONTEXT ctx)
{
  BOOL direct = !(ctx->flags & FLAG_CINDIRECT);
  int (*WineXHandler)(Display *, XErrorEvent *);
  HRESULT res = D3D_OK;
  LPD3DCONTEXT old_ctx = X11DRV_GLX_GetCurrent();
  pid_t our_pid = getpid();

  if (ctx == old_ctx)
    return D3D_OK; /* already current */

  if (old_ctx) X11DRV_GLX_KillCurrent(old_ctx);
  if (ctx->owner_pid && ctx->owner_pid != our_pid) {
    /* invoke owner thread's X11DRV_GLX_UserSignal */
    kill(ctx->owner_pid, SIGUSR1);
    while (ctx->owner_pid) usleep(1000);
  }

  ENTER_GL();
  ctx->save_ctx = glXGetCurrentContext();
  ctx->save_draw = glXGetCurrentDrawable();
  if (ctx->ctx) {
    /* hmm, just re-establish current context? */
    BOOL ok = glXMakeCurrent(gdi_display, ctx->draw, ctx->ctx);
    TRACE("resetting context %p, ctx %p, bmp %08x, draw %08lx\n", ctx, ctx->ctx, ctx->bmp, ctx->draw);
    if (!ok) {
      ERR("failed to reset OpenGL context\n");
      res = D3DERR_INVALID_DEVICE;
    }
  }
  while (!ctx->ctx) {
    BOOL ok;
    ctx->ctx = glXCreateContext(gdi_display, ctx->vis, 0, direct);
    TRACE("context %p, ctx %p, bmp %08x, draw %08lx, %s\n", ctx, ctx->ctx, ctx->bmp, ctx->draw, direct?"direct":"indirect");
    WineXHandler = XSetErrorHandler(GLXErrorHandler);
    GLXError = 0;
    ok = glXMakeCurrent(gdi_display, ctx->draw, ctx->ctx);
    XSetErrorHandler(WineXHandler);
    if (GLXError) {
      glXDestroyContext(gdi_display, ctx->ctx);
      ctx->ctx = 0;
      if (direct) {
        ERR("direct rendering to GLXPixmap rejected by OpenGL implementation, switching to indirect rendering\n");
        direct = FALSE;
      } else {
        ERR("rendering to GLXPixmap rejected by OpenGL implementation\n");
        res = D3DERR_INVALID_DEVICE;
        break;
      }
    }
    else if (!ok) {
      glXDestroyContext(gdi_display, ctx->ctx);
      ctx->ctx = 0;
      ERR("failed to set OpenGL context\n");
      res = D3DERR_INVALID_DEVICE;
      break;
    }
  }
  if (res == D3D_OK) {
    TRACE("initializing context\n");
    /* success - update state to reflect new render target */
    ctx->lsurf = ctx->nsurf;
    ctx->surf = ctx->lsurf ? ctx->lsurf->lpGbl : (LPDDRAWI_DDRAWSURFACE_GBL)NULL;
    /* cache results of direct/indirect check */
    if (direct) {
      if (ctx->flags & FLAG_PIXMAP) ctx->flags |= FLAG_PDIRECT;
      else ctx->flags |= FLAG_WDIRECT;
    } else {
      if (ctx->flags & FLAG_PIXMAP) ctx->flags |= FLAG_PINDIRECT;
      else ctx->flags |= FLAG_WINDIRECT;
    }
    /* choose front or back buffer for rendering */
    if (ctx->flags & FLAG_BACKBUF) {
      glDrawBuffer(GL_BACK);
      TRACE("rendering to GL_BACK\n");
    } else {
      glDrawBuffer(GL_FRONT);
      TRACE("rendering to GL_FRONT\n");
    }
    ctx->owner_pid = our_pid;
    X11DRV_GLX_SetCurrent(ctx);
  } else {
    /* failure - revert state to old render target */
    TRACE("resetting context\n");
    if (ctx->lsurf && ctx->lsurf != ctx->nsurf)
      X11DRV_GLX_MakeTarget(ctx, ctx->lsurf);
  }

  LEAVE_GL();
  return res;
}

static void X11DRV_GLX_DoKillCurrent(LPD3DCONTEXT ctx)
{
  glXMakeCurrent(gdi_display, ctx->save_draw, ctx->save_ctx);
  ctx->owner_pid = 0;
  X11DRV_GLX_SetCurrent(NULL);
}

static void X11DRV_GLX_KillCurrent(LPD3DCONTEXT ctx)
{
  if (ctx != X11DRV_GLX_GetCurrent())
    return;

  ENTER_GL();
  X11DRV_GLX_DoKillCurrent(ctx);
  LEAVE_GL();
}

void X11DRV_GLX_UserSignal(CONTEXT* pctx)
{
  LPD3DCONTEXT ctx = X11DRV_GLX_GetCurrent();
  if (ctx) {
    WARN("deselecting GLX context from signal handler - can be risky\n");
    /* the X11/GL lock is held by the calling thread */
    X11DRV_GLX_DoKillCurrent(ctx);
  }
}

void X11DRV_GLX_Detach(void)
{
  LPD3DCONTEXT ctx = X11DRV_GLX_GetCurrent();
  if (ctx) X11DRV_GLX_KillCurrent(ctx);
}

static DWORD PASCAL X11DRV_GLX_ContextCreate(LPD3DHAL_CONTEXTCREATEDATA data)
{
  int num;
  XVisualInfo template;
  LPD3DCONTEXT ctx = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizeof(D3DCONTEXT));
  const char *exts;
  HRESULT res;

  TRACE("creating 3D context %p\n", ctx);
  /* FIXME: do this properly */
  USER_Lock();
  TRACE("locking render target\n");
  X11DRV_GLX_LockTarget(data->u2.lpDDSLcl, TRUE);
  ENTER_GL();
  template.visualid = XVisualIDFromVisual(visual);
  ctx->vis = XGetVisualInfo(gdi_display, VisualIDMask, &template, &num);
  TRACE("preparing render target\n");
  X11DRV_GLX_MakeTarget(ctx, data->u2.lpDDSLcl);
  res = X11DRV_GLX_MakeCurrent(ctx);

  if (res == D3D_OK) {
    TRACE("preparing Z buffer\n");
    X11DRV_GLX_MakeZbuffer(ctx, data->u3.lpDDSZLcl);

#ifdef HAVE_GLX_GETPROCADDRESS
    /* check GL and GLX extensions */
    exts = glGetString(GL_EXTENSIONS);
    if (strstr(exts, "GL_ARB_multitexture")) {
      void *func = glXGetProcAddressARB("glActiveTextureARB");
      ctx->glActiveTextureARB = func;
      glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &ctx->texunits);
    }

    X11DRV_GLX_EB_Context_Init();
#endif

    X11DRV_GLX_SetupLogo(ctx);
  }

  LEAVE_GL();

  if (res == D3D_OK) {
    TRACE("created 3D context\n");
  } else {
    ERR("error creating 3D context\n");
    ENTER_GL();
    X11DRV_GLX_Destroy(ctx);
    LEAVE_GL();
    HeapFree(GetProcessHeap(),0,ctx);
    ctx = NULL;
  }
  X11DRV_GLX_LockTarget(data->u2.lpDDSLcl, FALSE);
  USER_Unlock();
  data->dwhContext = (DWORD)ctx;
  data->ddrval = res;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_GLX_ContextDestroy(LPD3DHAL_CONTEXTDESTROYDATA data)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->dwhContext;

  TRACE("destroyed 3D context %p\n", ctx);

  if (ctx) {
    X11DRV_GLX_LockTarget(ctx->lsurf, TRUE);
    ENTER_GL();
    X11DRV_GLX_Destroy(ctx);
    LEAVE_GL();
    X11DRV_GLX_LockTarget(ctx->lsurf, FALSE);
    HeapFree(GetProcessHeap(),0,ctx);
  }

  data->ddrval = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_GLX_ContextDestroyAll(LPD3DHAL_CONTEXTDESTROYALLDATA data)
{
  /* since this isn't a real display driver, we probably don't need to do anything here,
   * the kernel and the X server will clean up after Wine exits */
  data->ddrval = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static void X11DRV_GLX_SetupLogo(LPD3DCONTEXT ctx)
{
#ifdef TRANSGAMING_LOGO
/* makedep is overzealous and doesn't obey the #ifdef. */
/*#include "logo.c"*/
#error "Uncomment the above line, and comment this one."

    ctx->logo_size = logo_size;

    glGenTextures(1, &ctx->logo);
    glBindTexture(GL_TEXTURE_2D, ctx->logo);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ctx->logo_size, ctx->logo_size,
		 0, GL_RGBA, GL_UNSIGNED_BYTE, logo);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);

    ctx->logo_list = glGenLists(1);

    glNewList(ctx->logo_list, GL_COMPILE);

    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT
		 | GL_DEPTH_BUFFER_BIT | GL_TEXTURE_BIT | GL_TRANSFORM_BIT
		 | GL_VIEWPORT_BIT);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0.0, ctx->logo_size, ctx->logo_size, 0.0, -1, 1);

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_SCISSOR_TEST);
    glMatrixMode(GL_TEXTURE);
    if (ctx->texunits > 1) {
	int i;
	for (i=0; i<ctx->texunits; i++) {
	    ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
	    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	    glDisable(GL_TEXTURE_2D);

	    glPushMatrix();
	    glLoadIdentity();
	    glDisable(GL_TEXTURE_GEN_S);
	    glDisable(GL_TEXTURE_GEN_T);
	    glDisable(GL_TEXTURE_GEN_R);
	    glDisable(GL_TEXTURE_GEN_Q);
	}
	ctx->glActiveTextureARB(GL_TEXTURE0_ARB);
    }
    else
    {
	glDisable(GL_TEXTURE_2D);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);
    }
    /* disable cubemapping, 3d */

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glDisable(GL_DEPTH_TEST);

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glColor4f(1, 1, 1, 0.333);
    if (gluCheckExtension("GL_EXT_secondary_color",
			  glGetString(GL_EXTENSIONS)))
    {
	glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
	glSecondaryColor3fEXT(0, 0, 0);
    }

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, ctx->logo);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT, 0);
    /* LOD bias shouldn't make a difference here. */

    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);	glVertex2f(0, 0);
    glTexCoord2f(1, 0);	glVertex2f(ctx->logo_size, 0);
    glTexCoord2f(1, 1);	glVertex2f(ctx->logo_size, ctx->logo_size);
    glTexCoord2f(0, 1);	glVertex2f(0, ctx->logo_size);
    glEnd();

    glMatrixMode(GL_TEXTURE);
    if (ctx->texunits > 1) {
	int i;
	for (i=0; i<ctx->texunits; i++) {
	    ctx->glActiveTextureARB(GL_TEXTURE0_ARB + i);
	    glPopMatrix();
	}
	ctx->glActiveTextureARB(GL_TEXTURE0_ARB);
    }
    else
    {
	glPopMatrix();
    }
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glPopAttrib();
    glPopClientAttrib();

    glEndList();
#endif
}

static void X11DRV_GLX_RenderLogo(LPD3DCONTEXT ctx)
{
#ifdef TRANSGAMING_LOGO
    glPushAttrib(GL_VIEWPORT_BIT);
    glViewport(ctx->lsurf->lpGbl->wWidth - ctx->logo_size,
	       ctx->lsurf->lpGbl->wHeight - ctx->logo_size,
	       ctx->logo_size, ctx->logo_size);

    glCallList(ctx->logo_list);

    glPopAttrib();
#endif
}

static DWORD PASCAL X11DRV_GLX_SceneCapture(LPD3DHAL_SCENECAPTUREDATA data)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->dwhContext;
  int s, sz;

  switch (data->dwFlag) {
  case D3DHAL_SCENE_CAPTURE_START:
    USER_Lock();
    X11DRV_GLX_MakeTarget(ctx, ctx->lsurf);
    TRACE("Locking bitmap %08x\n", ctx->bmp);
    ctx->lockedBmp = (BITMAPOBJ *)GDI_GetObjPtr(ctx->bmp, BITMAP_MAGIC);
    s = X11DRV_DIB_Lock(ctx->lockedBmp, (ctx->flags & FLAG_PIXMAP) ? DIB_Status_GdiMod : DIB_Status_AuxMod, FALSE);
    if (ctx->zbmp) {
      ctx->lockedZBmp = (BITMAPOBJ *)GDI_GetObjPtr(ctx->zbmp, BITMAP_MAGIC);
      sz = X11DRV_DIB_Lock(ctx->lockedZBmp, DIB_Status_AuxMod, FALSE);
      GDI_ReleaseObj(ctx->zbmp);
    }
    GDI_ReleaseObj(ctx->bmp);
    USER_Unlock();
    ENTER_GL();
    glXWaitX();
    X11DRV_GLX_MakeCurrent(ctx);
#ifdef RENDER_TO_WINDOW
    if ((s != DIB_Status_AuxMod) && !(ctx->flags & FLAG_PIXMAP)) X11DRV_GLX_CopyDIB(ctx, FALSE);
#endif
    if (ctx->zbmp && sz != DIB_Status_AuxMod) X11DRV_GLX_CopyDIB(ctx, TRUE);
    break;
  case D3DHAL_SCENE_CAPTURE_END:
    X11DRV_GLX_RenderLogo(ctx);
    /* glFinish(); */
    glXWaitGL();
    LEAVE_GL();
    TRACE("Unlocking bitmap %08x\n", ctx->bmp);
    if (ctx->zbmp) {
      X11DRV_DIB_Unlock(ctx->lockedZBmp, TRUE);
      ctx->lockedZBmp = NULL;
    }
    X11DRV_DIB_Unlock(ctx->lockedBmp, TRUE);
    ctx->lockedBmp = NULL;
    break;
  }
  data->ddrval = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

static DWORD PASCAL X11DRV_GLX_SetRenderTarget(LPD3DHAL_SETRENDERTARGETDATA data)
{
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->dwhContext;
  HRESULT res = D3DERR_INVALID_DEVICE;

  if (ctx) {
    ENTER_GL();
    X11DRV_GLX_MakeTarget(ctx, data->u1.lpDDSLcl);
    res = X11DRV_GLX_MakeCurrent(ctx);
    LEAVE_GL();
    if (res == D3D_OK) {
      X11DRV_GLX_MakeZbuffer(ctx, data->u2.lpDDSZLcl);
    }
  }

  data->ddrval = res;
  return DDHAL_DRIVER_HANDLED;
}

static inline void X11DRV_GLX_SetContext(LPD3DCONTEXT ctx)
{
  if (ctx->lockedBmp) return; /* locked => already active */
  /* set context */
  X11DRV_GLX_MakeCurrent(ctx);
}

static LPDDHAL_CREATESURFACE X11DRV_GLX_old_create_surface;
static LPDDHALSURFCB_DESTROYSURFACE X11DRV_GLX_old_destroy_surface;
static LPDDHALSURFCB_FLIP X11DRV_GLX_old_flip;

static DWORD PASCAL X11DRV_GLX_CreateSurface(LPDDHAL_CREATESURFACEDATA data)
{
  LPDDRAWI_DDRAWSURFACE_LCL lcl = *data->lplpSList;
  LPDDRAWI_DDRAWSURFACE_GBL gbl = lcl->lpGbl;
  LPDDRAWI_DDRAWSURFACE_GBL_MORE gblm = GET_LPDDRAWSURFACE_GBL_MORE(gbl);
  LPDDSURFACEDESC2 desc = (LPDDSURFACEDESC2)data->lpDDSurfaceDesc;
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)lcl->dwReserved1;

  if (desc->ddsCaps.dwCaps & DDSCAPS_TEXTURE) {
    /* only process late allocations (with a context) */
    if (ctx && !(desc->ddsCaps.dwCaps2 & DDSCAPS2_MIPMAPSUBLEVEL)) {
      ENTER_GL();
      X11DRV_GLX_SetContext(ctx);
      glGenTextures(1, (GLuint*)&gblm->dwDriverReserved);
      LEAVE_GL();
      TRACE("generated texture %d\n", gblm->dwDriverReserved);
    }
  }

  if (X11DRV_GLX_old_create_surface)
    return X11DRV_GLX_old_create_surface(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_GLX_DestroySurface(LPDDHAL_DESTROYSURFACEDATA data)
{
  LPDDRAWI_DDRAWSURFACE_LCL lcl = data->lpDDSurface;
  LPDDRAWI_DDRAWSURFACE_MORE lclm = lcl->lpSurfMore;
  LPDDRAWI_DDRAWSURFACE_GBL gbl = lcl->lpGbl;
  LPDDRAWI_DDRAWSURFACE_GBL_MORE gblm = GET_LPDDRAWSURFACE_GBL_MORE(gbl);
  LPD3DCONTEXT ctx = (LPD3DCONTEXT)lcl->dwReserved1;

  if (lcl->ddsCaps.dwCaps & DDSCAPS_TEXTURE) {
    if (ctx && gblm->dwDriverReserved) {
      TRACE("deleted texture %d\n", gblm->dwDriverReserved);
      ENTER_GL();
      if (!ctx->lockedBmp) {
        X11DRV_GLX_SetContext(ctx);
        glXWaitX();
      }
      glDeleteTextures(1, (GLuint*)&gblm->dwDriverReserved);
      gblm->dwDriverReserved = 0;
      if (!ctx->lockedBmp) glXWaitGL();
      LEAVE_GL();
    }
  }

  if (X11DRV_GLX_old_destroy_surface)
    return X11DRV_GLX_old_destroy_surface(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}

static DWORD PASCAL X11DRV_GLX_Flip(LPDDHAL_FLIPDATA data)
{
  if (data->lpSurfCurr == X11DRV_DD_Primary) {
    LPD3DCONTEXT ctx = (LPD3DCONTEXT)data->lpSurfTarg->dwReserved1;
    TRACE("front=%p, back=%p, ctx=%p\n", data->lpSurfCurr, data->lpSurfTarg, ctx);
    if (ctx && ctx->lsurf == data->lpSurfTarg &&
        !(ctx->flags & FLAG_PIXMAP)) {
      /* we have a windowed render target */
      DWORD done = FALSE;
      int state;

      if (ctx->lockedBmp) {
        /* don't know if this could cause problems or not */
        TRACE("calling Flip on surface locked for rendering\n");
        state = ctx->dib ? ctx->dib->status : DIB_Status_None;
      }
      else state = X11DRV_LockDIBSection2(ctx->bmp, DIB_Status_None, FALSE);

      if (state == DIB_Status_AuxMod) {
        TRACE("performing GLX flip\n");
        ENTER_GL();
        X11DRV_GLX_SetContext(ctx);
        glXSwapBuffers(gdi_display, ctx->draw);
        LEAVE_GL();
        ctx->flags ^= FLAG_FLIPPED;
        done = TRUE;
      }
      else {
        TRACE("state=%d, not flipping\n", state);
      }

      if (!ctx->lockedBmp) X11DRV_UnlockDIBSection2(ctx->bmp, done);
      if (done) return DDHAL_DRIVER_HANDLED;
    }
  }
  if (X11DRV_GLX_old_flip)
    return X11DRV_GLX_old_flip(data);
  else
    return DDHAL_DRIVER_NOTHANDLED;
}


static D3DHAL_CALLBACKS d3dhal_callbacks = {
  sizeof(D3DHAL_CALLBACKS),
  X11DRV_GLX_ContextCreate,
  X11DRV_GLX_ContextDestroy,
  X11DRV_GLX_ContextDestroyAll,
  X11DRV_GLX_SceneCapture,
  /* use lpReserved10 for something useful, like GetProcAddress */
#ifdef HAVE_GLX_GETPROCADDRESS
  glXGetProcAddressARB
#else
  NULL
#endif
};

static D3DHAL_CALLBACKS2 d3dhal_callbacks2 = {
  sizeof(D3DHAL_CALLBACKS2),
  0x1f, /* all callbacks are 32-bit */
  X11DRV_GLX_SetRenderTarget
};

static D3DHAL_GLOBALDRIVERDATA d3dhal_driverdata = {
  sizeof(D3DHAL_GLOBALDRIVERDATA),
  {	/* hwCaps */
   sizeof(D3DDEVICEDESC_V1),
   D3DDD_WINE_OPENGL_DEVICE
  },
  0,	/* dwNumVertices */
  0,	/* dwNumClipVertices */
  0,	/* dwNumTextureFormats */
  NULL	/* lpTextureFormats */
};

static X11DRIVERINFO driver_info[] = {
  {&GUID_D3DCallbacks2, sizeof(D3DHAL_CALLBACKS2), &d3dhal_callbacks2, NULL},
  {&GUID_WineGLD3DCallbacks, sizeof(struct GL_D3D_DriverCallbacks),
   &X11DRV_GLX_GL_D3D_driver_callbacks, NULL}
};

int X11DRV_GLX_CreateDriver(LPDDHALINFO info)
{
  LPX11DEVICE device = (LPX11DEVICE)info->lpPDevice;

  if (!glx_available) return 0;

  /* attach Direct3D callbacks */
  info->lpD3DGlobalDriverData = (ULONG_PTR)&d3dhal_driverdata;
  info->lpD3DHALCallbacks = (ULONG_PTR)&d3dhal_callbacks;
  info->lpDDExeBufCallbacks = &X11DRV_GLX_exebufcallbacks;

  /* add 3D caps bits */
  info->ddCaps.dwCaps |= DDCAPS_3D;
  info->ddCaps.dwCaps2 |= DDCAPS2_NO2DDURING3DSCENE;
  info->ddCaps.ddsCaps.dwCaps |= DDSCAPS_TEXTURE | DDSCAPS_3DDEVICE |
                                 DDSCAPS_ZBUFFER | DDSCAPS_MIPMAP |
                                 DDSCAPS_EXECUTEBUFFER | DDSCAPS_OPTIMIZED;

  /* attach our D3D callback structures to driver info linked list */
  driver_info[0].lpNext = driver_info+1;
  driver_info[1].lpNext = device->lpInfo;
  device->lpInfo = driver_info;

  /* chain some surface callbacks */
  X11DRV_GLX_old_create_surface = info->lpDDCallbacks->CreateSurface;
  X11DRV_GLX_old_destroy_surface = info->lpDDSurfaceCallbacks->DestroySurface;
  X11DRV_GLX_old_flip = info->lpDDSurfaceCallbacks->Flip;
  info->lpDDCallbacks->CreateSurface = X11DRV_GLX_CreateSurface;
  info->lpDDSurfaceCallbacks->DestroySurface = X11DRV_GLX_DestroySurface;
  info->lpDDSurfaceCallbacks->Flip = X11DRV_GLX_Flip;

  return TRUE;
}
